#!/usr/bin/perl
# dungeon2.pl [name=<name>] [up=<up>]* [down=<down>]* [size=<size>] (<monster>=<count>)*
# generate dungeon style levels for hacknet

#Copyright 2002 Peter Hanely, hanelyp@calweb.com

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


sub set_tile
{ my ($x, $y, $c) = @_;
  $line = $base[$y];
  if (chkbnd($x, $y) && (substr($line, $x, 1) ne $room))
  { substr($line, $x, 1) = $c;
    $base[$y] = $line;
#print '+';
#print "($x,$y)$c\n";
  }
#  else
#  { print "error in bound ($x, $y) : $c\n"; }
}

sub chkbnd
{ my ($x, $y) = @_;
  ($x >=0) && ($x < $size) && ($y >= 0) && ($y < $size );
}

sub generate_passage
{ my ($cx, $cy, $sx, $sy) = @_;
  print '.' if $watchit;
  $ex = $ey = -1;
  $trycount = 0;
  while (!chkbnd($ex, $ey) && ($trycount++ < 8))
  { $dir = 1 + int rand (4);
    if ($dir == 1)
    { $ex = $cx + $sx + int rand 2;
      $ey = $cy + int (-$sy/3 + rand (0.67*$sy));
      if (&chkbnd ($ex, $ey))
      { for ($x = $cx; $x <= $ex; $x ++)
        { &set_tile($x, $cy, $passage);}
      }
    }
    elsif ($dir == 2)
    { $ey = $cy + $sy + int rand 2;
      $ex = $cx + int (-$sx/3 + rand (0.67*$sx));
      if (&chkbnd ($ex, $ey))
      { for ($y = $cy; $y <= $ey; $y ++)
        { &set_tile($ex, $y, $passage);}
      }
    }
    elsif ($dir == 3)
    { $ex = $cx - $sx - int rand 2;
      $ey = $cy + int (-$sy/3 + rand (0.67*$sy));
      if (&chkbnd ($ex, $ey))
      { for ($x = $cx; $x >= $ex; $x --)
        { &set_tile($x, $cy, $passage);}
      }
    }
    elsif ($dir == 4)
    { $ey = $cy - $sy - int rand 2;
      $ex = $cx + int (-$sx/3 + rand (0.67*$sx));
      if (&chkbnd ($ex, $ey))
      { for ($y = $cy; $y >= $ey; $y --)
        { &set_tile($ex, $y, $passage);}
      }
    }
  }
  push @passages, "$ex:$ey:$dir";
}

sub generate_room
{ my ($psg, $clean) = shift @_; # where to build
#  print '|';
  $thisroom = $clean || $room;
  # center room off end of passage
  ($cx, $cy, $dir) = split ':', $psg;
#  print "building $psg($cx,$cy)$dir\n";

  $clear = $trys = $spawning = 0;

  while (!($clear) && (($trys++) <8))
  {
    $clear = 1;
    $xs = 3 + int rand(3);
    $ys = 3 + int rand(3);

    if ($dir == 1)
    { $cx += int($xs/2);
      $cy = $cy - int ($ys/3 + rand(0.67*$ys)); 
    }
    elsif ($dir == 2)
    { $cy += int($ys/2);
      $cx = $cx - int ($xs/3 + rand(0.67*$xs)); 
    } 
    elsif ($dir == 3)
    { $cx -= int($xs/2);
      $cy = $cy - int ($ys/3 + rand(0.67*$ys)); 
    } 
    elsif ($dir == 4)
    { $cy -= int($ys/2);
      $cx = $cx - int ($xs/3 + rand(0.67*$xs)); 
    } 

    $x1 = $cx - int($xs/2); $x2 = $cx + int($xs/2);
    $y1 = $cy - int($ys/2); $y2 = $cy + int($ys/2);
    
    if (!&chkbnd($cx, $cy)) { $clear = 0; } # next;
  # check if area is clear
    elsif (!&chkbnd($x1,$y1) || !&chkbnd($x2,$y2))
    { $ clear = 0; }
  #print "($x1,$y1)($x2,$y2)\n";
    else 
    {for ($x = $x1 - 1; $x <= $x2 + 1; $x ++)
     { for ($y = $y1 - 1; $y <= $y2 + 1; $y ++)
      { $spot = substr($base[$y], $x, 1);
        if (($spot eq $room) || ($spot eq $startroom))
        { $clear = 0; }
      }
     }
    }
  } # while trying to find clear spot
#  print '/';
  # if so, dig room, 
  if ($clear != 0)
  { print "room ($cx,$cy) ($xs x $ys)\n" if $watchit;
    for ($x = $x1; $x <= $x2; $x ++)
    { for ($y = $y1; $y <= $y2; $y ++)
      { &set_tile ($x, $y, $thisroom);
    #print "($x,$y)$room\n";
      }
    }
#    $cx1 = $cx * 64;  $cy1 = $cy * 64;
#    push @entities, "SPAWN $cx1 $cy1 128 0\n";
#  print '-';  
  # also dig some outgoing passages
    &generate_passage($cx,$cy, $xs, $ys);
    &generate_passage($cx,$cy, $xs, $ys);
    &generate_passage($cx,$cy, $xs, $ys);
  } # if clear
}

$solid = '|';
$room = '.';
$startroom = '.';
$passage = '#';
# a few defaults
$size = 50;

select STDERR;
$| = 1;

print "reading arguments...\n";
while ($arg = shift @ARGV)
{ ($name, $val) = split '=', $arg, 2;
  if ($name eq "size")
  { $size = $val || $size; }
  elsif ($name eq 'up')
  { push @up, $val; }
  elsif ($name eq 'down')
  { push @down, $val; }
  elsif ($name eq'name')
  { $levelname=$val; }
  else # a monster
  { $monsters{"\t$name"} = "\t$val";
  }
}

if (!%monsters)
{ %monsters = ("monster\tTroll"=>$size*$size/250, "\tScooby Snack"=>$size*$size/1500);
} # a few basic hostiles

print "establishing ground...\n";  

for ($i = 0; $i < $size; $i ++)
{ $bline .= $solid;
}
$bline .= "\n";
for ($i = 0; $i < $size; $i ++)
{ push @base, $bline;
}

print "excavating";
# generate rooms and passages
# init dungeon ground
#  $cx = $xs + int rand ($size - 4*$xs);
#  $cy = $ys + int rand ($size - 4*$xs);
#$cx = 3+ int rand 3;
#$cy = 6+ int rand 3;
$cx = $cy = int ($size/2);

&generate_room("$cx:$cy:1", $startroom);

while ($psg = shift @passages)
{ &generate_room($psg); }

print "\npopulating...\n";
# populate with critters
foreach $monster(keys %monsters)
{ print "$monster\n";
  $count = $monsters{$monster};
  for ($i = 0; $i < $count;)
  { $x = int rand $size;
    $y = int rand $size;
    if (substr($base[$y], $x, 1) eq $room)
    { substr($base[$y], $x, 1) = $passage; # mark as occupied
      push @entities, "$monster\t$x,$y\t60\t$x,$y\n";
      $i ++;
    }
  }
} # each monster type

#stairs, up and down
foreach $up (@up)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tupstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCH:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x) .','.($y1+$y);  # so incoming stair can find location
	$okstair ++;
	last SEARCH;
      }
    }
  }
  if (!$okstair) { print STDERR "BAD STAIR!!!\n\n"; }
}
foreach $up (@down)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tdownstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCHA:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x).','.($y1+$y);  # so incoming stair can find location
        $okstair ++;
	last SEARCHA;
      }
    }
  }
}



select STDOUT;
$| = 1;
# output.  pipe into file
print "level=$levelname\n$size x $size\n",
	"dungeon2.pl, Peter Hanely\n",
	join ("\n", @head),
	"\n[MAP]\n", @base;
#for ($i = 0; $i < $size; $i ++)
#{ print $oline;}
print "\n\n", @entities;
	
