#!/usr/bin/perl
# forest.pl [size=<size>]
# variation on dungeon.pl with round rooms and radial paths

#Copyright 2002 Peter Hanely, hanelyp@calweb.com

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$pi = 3.1415926;

sub set_tile
{ my ($x, $y, $c) = @_;
  $x = int $x;  $y = int $y;
  $line = $base[$y];
  if (chkbnd($x, $y) && (substr($line, $x, 1) ne $room))
  { substr($base[$y], $x, 1) = $c;
  }
}

#used?
#sub draw_path
#{ my ($x1, $y1, $y1, $y2, $c) = @_;
#  my ($dx, $dy) = ($x2-$x1, $y2-$y1);
#  $dist = sqrt($dx*$dx+$dy*$dy);
#}

sub chkbnd
{ my ($x, $y) = @_;
  ($x >=0) && ($x < $size) && ($y >= 0) && ($y < $size );
}

sub generate_passage
{ my ($cx, $cy, $rad, $dir) = @_;
  my ($dx,$dy, $ex,$ey);
  print '.' if $watchit;
  $ex = $ey = -1;
  $trycount = 0;
  while (!chkbnd($ex, $ey) && ($trycount++ < 8))
  { #$dir = rand (360);
    $rad1 = $rad + rand ($rad + 6);
    $ex = int ($cx + $rad1 * cos($dir * $pi/180));
    $ey = int ($cy + $rad1 * sin($dir * $pi/180));
  }
  if (chkbnd($ex,$ey))
  { push @passages, "$ex:$ey";
#    &drawpath($cx,$cy, $ex,$ey, $passage);
    # short interval is attempt at passable path on diagonals
    for ($r = 0; $r <= $rad1; $r += 0.1)
    { $dx = $cx + $r * cos ($dir * $pi/180);
      $dy = $cy + $r * sin ($dir * $pi/180);
      #print "path $dx,$dy\n";
      &set_tile($dx,$dy, $passage);
      &set_tile($dx+0.2,$dy+0.2, $passage);
    }
    #print "passage ($cx,$cy)($ex,$ey)$dir\n";
  }
}

sub generate_room
{ my ($psg) = shift @_; # where to build
#  print '|';
  print "\nbuilding $psg\n" if $watchit;

  $clear  = $spawning = 0;

#  $clear = 1;
  if ($psg =~ /(\d+):(\d+)/)
  { # center room on end of passage
    $cx = $1; $cy = $2;
  }
  else
  { # no starting passage given, put it wherever
    #$cx = 6 + int rand ($size - 12);
    #$cy = 6 + int rand ($size - 12);
    $cx = $cy = $size/2;
    # also record a spawning point here
    $spawning = 1;
  }
  $trycount = 0;
  while ((!$clear) && ($trycount++ < 8))
  { $clear = 1;
    $rad = 1.5 + rand(1.5);
  # check if area is clear
  #print "($x1,$y1)($x2,$y2)\n";
  for ($dir = 0; $dir <= 2*$pi; $dir += $pi/$rad)
  { for ($r = 0; $r <= ($rad+2); $r += 0.2)
    { $spot = substr($base[int $cy + $r * sin $dir],
		int $cx + $r * cos $dir, 1);
      if (($spot eq $room) || ($spot eq ''))
      { $clear = 0; }
    }
  }
  }
#  print '/';
#  print "room ($cx,$cy)*$rad\n";
  # if so, dig room, 
  if ($clear != 0)
  { for ($dir = 0; $dir <= 2*$pi; $dir += $pi/$rad/3)
    { for ($r = 0; $r <= $rad; $r += 0.2)
      { $dx = $cx + $r * cos $dir;
        $dy = $cy + $r * sin $dir;
        &set_tile($dx, $dy, $room);
      }
    }
  }
#  if ($spawning)
#  { $startx = $cx; $starty = $cy;
#    $cx1 = $cx * 64;  $cy1 = $cy * 64;
#  }
#  print '-';  
  # also dig some outgoing passages
  if ($clear)
  { $dir = rand (360);
    &generate_passage($cx,$cy, $rad+rand(6), $dir);
    $dir += 60 + rand(120);
    &generate_passage($cx,$cy, $rad+rand(6), );
    $dir += 60 + rand(120);
    &generate_passage($cx,$cy, $rad+rand(6), $dir);
#    &generate_passage($cx,$cy, $rad);
  }
}

$solid = '|';
$room = '.';
$passage = '#';
# a few defaults
$size = 50;

select STDERR;
$| = 1;

print "reading arguments...\n";
while ($arg = shift @ARGV)
{ ($name, $val) = split '=', $arg, 2;
  if ($name eq "size")
  { $size = $val || $size; }
    elsif ($name eq 'up')
  { push @up, $val; }
  elsif ($name eq 'down')
  { push @down, $val; }
  elsif ($name eq'name')
  { $levelname=$val; }
    else # a monster
  { $monsters{"\t$name"} = "\t$val";
  }
}

if (!%monsters)
{ %monsters = ("monster\tBat"=>$size*$size/500, "\tquaffachino"=>$size*$size/1500);
} # a few basic hostiles

print "establishing ground...\n";  
# init dungeon ground
for ($i = 0; $i < $size; $i ++)
{ $bline .= $solid;
}
$bline .= "\n";
for ($i = 0; $i < $size; $i ++)
{ push @base, $bline;
}

print "excavating";
# generate rooms and passages
&generate_room();

while ($psg = shift @passages)
{ &generate_room($psg); }

print "\npopulating...\n";
# populate with critters
foreach $monster(keys %monsters)
{ print "$monster\n";
  $count = $monsters{$monster};
  for ($i = 0; $i < $count;)
  { $x = int rand $size;
    $y = int rand $size;
    if (substr($base[$y], $x, 1) eq $room)
    { substr($base[$y], $x, 1) = $passage; # mark as occupied
      push @entities, "$monster\t$x,$y\t60\t$x,$y\n";
      $i ++;
    }
  }
} # each monster type

#stairs, up and down
foreach $up (@up)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tupstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCH:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x) .','.($y1+$y);  # so incoming stair can find location
	$okstair ++;
	last SEARCH;
      }
    }
  }
  if (!$okstair) { print STDERR "BAD STAIR!!!\n\n"; }
}
foreach $up (@down)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tdownstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCHA:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x).','.($y1+$y);  # so incoming stair can find location
        $okstair ++;
	last SEARCHA;
      }
    }
  }
}

select STDOUT;
$| = 1;
# output.  pipe into file
print 	"level=$levelname\nsize=$size,$size\n",
	"forest.pl, Peter Hanely\n",
	join ("\n", @head),
	"\n[MAP]\n",
	@base, "\n";
print "\n\n", @entities;
