#!/usr/bin/perl
# hacknet.pl
# a networked roguelike

print <<license
Copyright 2002 Peter Hanely, hanelyp\@calweb.com

        This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
license
;
# note: show this to coworker in reference to obfuscated code.
# nested hashes and arrays can be tricky to read

require "tnet.pl";
require "ansivt.pl";
require "hnwizard.pl";
require "redraw.pl";
require "login.pl";

use watchscaler;

# ===========================
# general configuration
# ===========================
$timeout = 30*60; # 30 minutes
$mapmaster = 'hacknet'; #
$port = 1067;
$connections = 5;	# how many to listen for?
$savetime = 24*30*60; # 1 day
$loginprompt = "HackNet Login (<user> <passwd><enter>)\r\n: ";
$maxmsg = 50;
$PvP = 1;
$hungerrate = 0.03;
$maxhunger = 50;
$exp_bonus = exp(1)/1000;  # double hit every 1000 exp

$| = 1; # don't buffer, messes up stuff.

use POSIX;
# =========================================
# lookup tables.  commands listed here
# =========================================
# %commands
# a hash of functions to impliment each recognised command
# trimmed extraction from:
# http://nethack.sourceforge.net/v340/Guidebook.html#_TOCentry_7
%commands = ('?'=>\&help, '/'=>\&looksym, '&'=>\&helpcmd,
	'7'=>\&movei,	'8'=>\&movei,	'9'=>\&movei,
	'4'=>\&movei,	'5'=>\&rest,	'6'=>\&movei,
	'1'=>\&movei,	'2'=>\&movei,	'3'=>\&movei,
# classic Hacknet directions
	'y'=>\&movei,	'k'=>\&movei,	'u'=>\&movei,
	'h'=>\&movei,	'.'=>\&rest,	'l'=>\&movei,
	'b'=>\&movei,	'j'=>\&movei,	'n'=>\&movei,
	
	'm'=>\&go, 'F'=>\&fight, '.'=>\&rest,
	"\x03"=>\&quit, 'd'=>\&drop, 'e'=>\&eat,
	'f'=>\&fire, 'g'=>\&get, 'i'=>\&inventory, 'q'=>\&quaff, 'z'=>\&zap,
	'Q'=>\&quiver, 'R'=>\&read, 'S'=>\&save,
	't'=>\&throw, 'w'=>\&wear, 'W'=>\&wear, 'r'=>\&unwear,
	':'=>\&look, #'l'=>\&look,
	','=>\&get, '$'=>\&gold, '"'=>\&chat, "'"=>\&chat,#"\n"=>\&chat,
	'o'=>\&who, 'p'=>\&page,
	"\xff"=>\&telnet);

# x,y offset for direction numbers
%directions = (
	'7'=>[-1,1],	'8'=>[0,1],	'9'=>[1,1],
	'4'=>[-1,0],	'5'=>[0,0],	'6'=>[1,0],
	'1'=>[-1,-1],	'2'=>[0,-1],	'3'=>[1,-1],
# or for those who prefer classic HackNet directions
	'y'=>[-1,1],	'k'=>[0,1],	'u'=>[1,1],
	'h'=>[-1,0],	'.'=>[0,0],	'l'=>[1,0],
	'b'=>[-1,-1],	'j'=>[0,-1],	'n'=>[1,-1]);
%revdir = (
	'-1,1'=>'7',	'0,1'=>'8',	'1,1'=>'9',
	'-1,0'=>'4',	'0,0'=>'5',	'1,0'=>'6',
	'-1,-1'=>'1',	'0,-1'=>'2',	'1,-1'=>'3');
$directionkeys = '12346789ykuhlbjn';

%arrows = (
	'7'=>'\\',	'8'=>'|',	'9'=>'/',
	'4'=>'-',	'5'=>'O',	'6'=>'-',
	'1'=>'/',	'2'=>'|',	'3'=>'\\');
	
# extra / override commands if wizard is set on player
# defined with wizard commands
#%wizard = ('b'=>\&build, 'm'=>\&monster, 'M'=>\&monsterstat, 'm'=>\&shutdown);

$helpmain = <<HELP
'?'  help   '/'  look\@sym   '&' helpcmd,
'7 8 9' move, attack if player/monster in way
'4   6'
'1 2 3'
*'m'  move, no fight    'F'  fight
'f'   fire dart/arrow   'Q'  select from quiver
't'   throw             'w'  wear   'r'  remove
'.''5' rest
'g'',' get   'd'  drop     'i'  inventory
*'$'  gold
'e'   eat    'q'  quaff    'z'  Zap a wand
'R'   read   ':'  look,  '"'  chat
'o'   who/on 'p'  page
^C    quit   'S'  save
(*) not implimented
HELP
;

$SIG{INT} = sub { $shutdown ++; };
END
{ close $incomingsock;
  foreach $player (@players)
  { select $player->{'socket'};
    print ansivt::home(0,22),"Server Closing\r\n\n";
    close $player->{'socket'};
  }
}

$SIG{HUP} = sub { print "going demon\n"; };

# ==========================
# Load data for game
# ==========================

$starttime = $lastsave = time();
sub loadall
{
print "loading data\n";
# load the master map
$level = 0;
open MASTER, $mapmaster;
$/ = "\n";
while ($l = <MASTER>)
{ print "$l";
  chomp $l;
  ($type, $file) = split '=', $l, 2;
  # multiple levels in submap files
  if ($type eq 'map')
  { &loadmap ($file); $level++;}
  # player file
  elsif ($type eq 'players')
  { &loadobjects (\%players, $file); }
  # monsters and such
  elsif ($type eq 'monsters')
  { &loadobjects (\%monsters, $file); }
  elsif ($type eq 'tiles')
  { &loadtiles (\%tiles, $file); }
  elsif ($type =~ /^#/) {} # comment
  elsif ($type eq 'code')
  { require "$file"; }
  else { print "unknown game spec, master file\n$l\n"; }
}
close MASTER;

# load command help
open HELP, "commands.hlp";
$/ = "\n\n";
while ($c = <HELP>)
{ ($index, $val) = split /\s/, $c, 2;
  $help{$index} = $val;
}
close HELP;
} # loadall

&loadall();

sub loadmap
{ my ($file) = @_;
  local (*FL);
  my ($inline, $m, $mtype, $mat, $x, $y, $respawn, $xtra, %monst, @map);
  my ($tile, @y, $index);
  my ($xxtra, $xk, $xv);
  $x = $y = 0;
  open (FL, $file);

  # read stats
  while ((!eof(FL)) && (($inline = <FL>) !~ /\[MAP\]/i))
  { $inline =~ s/\n//;
    ($x, $y) = split '=', $inline;
    $levelstat[$level]->{$x} = $y;
    $index = $index || $y;
  }

  # read the map
  while ((!eof(FL)) && (($inline=<FL>) =~ /\S/))
  { chomp $inline;
    $y = [];
    while ($tile = substr($inline, 0,1))
    { push @$y, [$tile];
      $inline = substr($inline, 1);
    }
    push (@map, $y);
  }
  push @levels, \@map;
  $mapindex{$index} = $level if $index;
  $mapindex{$level} = $level;
  $maprevindx{$level} = $index || $level;
  
  # read in 'monsters'
  while ((!eof(FL)) && ($inline = <FL>))
  { chomp $inline;
    next if ($inline !~ /\w/);
    ($m, $mtype, $mat, $respawn, $respat, $xtra, $xxtra) = split /\t/, $inline, 7;
    ($x, $y, $count) = split ',', $mat;
    ($startx, $starty, $rscount) = split ',', $respat;
    $startx = ($startx || $x);
    $starty = ($starty || $y);
    $rscount = ($rscount || $count);
    $monst = {'type'=>$m, 'x'=>$x, 'y'=>$y, 'count'=>$count,
	'respawn'=>$respawn, 'xtra'=>$xtra, 'level'=>$level,
	'startx'=>$startx, 'starty'=>$starty,
	'startcount'=>$rscount, 'startmap'=>$level};
    foreach $key (keys (%{$monsters{$mtype}}) )
    { $monst->{$key} = $monsters{$mtype}->{$key} if ($key !~ /^[_|\*]/); }
    #if ($xxtra ne '')
    while ($xxtra !~ /^\s*$/)
    { foreach $m (split /\t/, $xxtra)
      { ($key, $xv) = split '=', $m;
        $monst->{$key} = $xv;
      }
      if ($key eq '\\') { $xxtra = <FL> ; chomp $xxtra; }
      else { $xxtra = ''; }
    }
    &insert_object ($monst, $level, $x, $y);
    #push @monsters, $monst; #monster list depreciated
  }  
  close FL;
} # sub loadmap

sub loadobjects
{ my ($hash, $file) = @_;
  local (*FL);
  my ($name, $val, $line, $index, %obj);
  my ($box, $boxlist, $boxitem);
  my ($key);
  open (FL, $file);
  while ((!eof(FL)) && (<FL> !~ /[OBJECT]/i)){}
  while (!eof(FL))
  { $obj = {};
    $index = "";
    while ((!eof(FL)) && (($line = <FL>) !~ /\[OBJECT\]/i) )
    { # print $line;
      chomp $line;
      next if ($line =~ /^#/);
      if ($line =~ /^%(\w+)/) # list within object
      { $box = $1;
        $boxlist = [];
        while ((!eof(FL)) && (($line = <FL>) ne "/\n") )
        { # print "$box-$line";
	  chomp $line;
	  ($name, $val) = split '=', $line, 2;
	  $boxitem = {'species'=>$name, 'count'=>$val};
	  push @$boxlist, $boxitem;
	  # copy attributes for this object type
	  foreach $key (keys %{$monsters{$name}})
	  { $boxitem->{$key} = $monsters{$name}->{$key} if ($key !~ /^_/); }
	}
	$obj->{$box} = $boxlist;
      }
      ($name, $val) = split '=', $line, 2;
      $index = $index || $val;

      if ($name eq 'isa')
      { foreach $key (keys (%{$monsters{$val}}) )
        { $obj->{$key} = $monsters{$val}->{$key} if ($key !~ /^_/); }
      }
      
      if ($val =~ /\\$/)
      { while ($val =~ /\\$/)
        { $val =~ s/\\$//;
          $obj->{$name} .= $val; #."\n";
	  $val = <FL>;
	}
	$obj->{$name} .= $val;
      }
      else { $obj->{$name} = $val; }
      if ($name =~ /^_(\w+)/)
      { $name1 = $1;
        $code = $obj->{$name};
        $code = " \$obj->{'$name1'} = sub{ my (\$monster, \$target, \$player) = \@_;\n$code\n};";
	eval ($code) || print "$code\n==========\n$@\n",;
      }
      if ($name =~ /^\&(\w+)/)
      { $name1 = $1;
	$code = $obj->{$name};
	$code = "\$obj->{'$name1'} = \\\&$code; ";
	eval $code;
	  print "$code\n====\n$@" if $@;
	  print ("unknown code $code\n") if ! ref $obj->{$name1};
      }
    }
    
    $hash->{$index} = $obj;
  }
  close FL;
}

sub loadtiles
{ my ($hash, $file) = @_;
  local (*FL);
  open FL, $file;
  while ($line = <FL>)
  { chomp $line;
    my ($tile, $stats, $descrip) = split "\t", $line, 3;
    my ($move, $visible, $dig, $color) = split ',', $stats, 4;
    $hash->{$tile} = {'move'=>$move, 'visable'=>$visible, 'dig'=>$dig,
	'color'=>$color, 'descrip'=>$descrip};
  }
  close FL;
}

# =================================================
# and finaly a network socket for connections
# =================================================
print "opening incoming socket\n";
$AF_INET = 2;
$SOCK_STREAM = 1;
$sockaddr = 'S n a4 x8';
($prname, $aliases, $proto) = getprotobyname('tcp');
$prname = $prname; $aliases = $aliases;

$this = pack($sockaddr, $AF_INET, $port, "\0\0\0\0");
socket($incomingsock, $AF_INET, $SOCK_STREAM, $proto) || die "socket: $!";
bind($incomingsock, $this) || die "bind: $!";
listen ($incomingsock, $connections);
#setsockopt $incomingsock, 
select ($incomingsock); $| = 1; select (STDOUT);
vec ($sockets, fileno($incomingsock), 1) = 1;

$SIG{'PIPE'} = 'IGNORE';


# ===============================================
# update player of anything effecting them
# ===============================================

# get command, interpret and update stats and maps, mark players
# limits to 3*speed moves a second
sub move
{ my ($player) = @_;
  my ($cmd, $goodcmd);
  #print "player:\n";
  #&dumpobj($player);
  while ($cmd = getc $player->{'socket'})
  { $goodcmd = 0;
#print STDERR "loop ", $player->{'name'}, "\n";
    if ($player->{'time'} != time())
    { $player->{'mvcnt'} = 0; }
    if ($cmd eq "\x03") { &quit($player, $cmd); next;} # try not to block an easy quit
    if (ref ($player->{'mode'}) )
    { $result = eval {&{$player->{'mode'}}($player, $cmd); }; 
      print "mode : $@" if $@;
      #print "mode\n";
      $goodcmd ++;
    }
    #{ &{$player->{'mode'}}($player, $cmd); }
    elsif (($player->{'wizard'}) && ($wizard{$cmd}))
    { $result = eval {&{$wizard{$cmd}}($player, $cmd); };
      $goodcmd ++;
    }
    elsif ($player->{'mvcnt'} > 3*($player->{'speed'} || 1))
    {}
    else
    { $result = eval {&{$commands{$cmd}}($player, $cmd); }; 
      $player->{'mvcnt'} ++;
      $goodcmd ++;
#print STDOUT "turn ", $player->{'name'}, "\n";
    }
    #if ($noact) { $noact = 0; }
    #else
    if ((!($player->{'mode'})) && ($goodcmd))
    { # queue player as target in nearby monsters
# print "yelling out\n";
      &notifyrangespeed($player, ' ') if (!$player->{'showmove'});
      # mark self for update
      $player->{'redraw'} = 1;
      $player->{'showmove'} = 0;
      $player->{'time'} = time();
      select ($player->{'socket'});
      print ansivt::home(2,21), ansivt::clreol();
      $player->{'hunger'} += $hungerrate;
      if ($player->{'hunger'} > $maxhunger)
      { $player->{'hunger'} = $maxhunger;
        $player->{'dammage'} += 1;
      }
      if (rand ($player->{'hunger'}/$maxhunger) > 0.67)
      { &notify ($player, 'your stomache growles'); }
    }
    $lastcommand = $cmd;
  }
}

# ==========================
# usefull routines
# ==========================
sub getstring
{ my ($player, $cmd) = @_;
  my ($line);

  while ($cmd ne "")
  { 
    if ($cmd =~ /[\n\r]/)
    { $line = $player->{'instring'};
      $line = 0 if (!$line);
      $player->{'instring'} = "";
      getc ($player->{'socket'}); 
      return $line;
    }
#    elsif ($cmd eq "\x03") { &quit($player, $cmd); print "GS abort\n"; last;} # try not to block an easy quit
    # elsif ($cmd eq "\x03") { return ""; }
    elsif ($cmd eq "") { return 0; }
    elsif ($cmd eq "\xff") {tnet::telnet ($player->{'socket'},$player); }
    elsif ((ord($cmd) > 127) || (ord($cmd) < 32)) {} # ignore
    else
    { $player->{'instring'} .= $cmd;
      select $player->{'socket'};
      print $cmd if ($player->{'ECHO'} ne 'DONT');
    }
    $cmd = getc $player->{'socket'};
  }
  '0';
}

# scans for objects in map
# $start	object or (map,x,y) tripple
# $direction	direction to scan or 0 for surround
# $distance	how far to scan
# $typespec	RE pattern, optional, to restrict return
# $count	how many are we intreested in?
sub findstuff
{ my ($start, $direction, $distance, $typespec, $count) = @_;
  my ($level, $x, $y, $spot, $obj, @list);
  my ($dx,$dy); # want $dist as output

  $typespec = $typespec || '.*';
  
  if (ref $start)
  { ($level, $x, $y) = ($start->{'level'}, $start->{'x'}, $start->{'y'}); }
  else
  { ($level, $x, $y) = split ',', $start; }
  
  if ($distance == 0)
  { eval { $spot = $levels[$level]->[$y]->[$x]; };
    if (defined $spot)
    { foreach $obj (@$spot)
      { push (@list, $obj) if (ref $obj); }
      return @list;
    }
  }
  if ($distance == 1) 
  { if ($direction)
    { ($x, $y) = ($x+$directions{$direction}->[0] , $y+$directions{$direction}->[1]);
      return (undef) if (($x < 0) || ($y < 0));
      eval { $spot = $levels[$level]->[$y]->[$x]; };
      if (defined $spot)
      { foreach $obj (@$spot)
        { push (@list, $obj) if (ref $obj); }
	return @list;
      }
    }
    else {  } # stub
  }
  else # distance > 1
  { if ($direction)
    { $dx = $directions{$direction}->[0];   $dy = $directions{$direction}->[1];
      for ($dist = 1; $dist <= $distance; $dist ++)
      { $x += $dx;  $y += $dy;
        $spot = '';
	eval { $spot = $levels[$level]->[$y]->[$x]; };
	last if (($tiles{$spot->[0]}->{'move'}) == 0);
	if ((ref $spot) && (@$spot > 1))
	{ 
	  foreach $obj (@$spot)
	  { push (@list, $obj) if (ref $obj) && ($obj->{'type'} =~ /$typespec/); }
	  return @list;
	}
      }
    }
    else #field search
    {}
  }
  undef; # default, config not yet written
} # sub findstuff

sub insert_object
{ my ($object, $level, $x, $y) = @_;
  my ($objtype, $spot, $at);

  return 0 if (($x < 0) || ($y < 0));
  return 1 if $object->{'norespawn'};
  $objtype = $object->{'type'};

  my ($key, $val, $package, $filename, $line);
  if (!$object->{'symbol'}) # bad object, dump debug
  { print "attempt to insert blank object\n";
    for ($key = 2; $key >= 0; $key--)
    { ($package, $filename, $line) = caller ($key);
      print "$filename: $line\n";
    }
    foreach $key (keys %$object)
    { $val = $object->{$key};
      print "$key => $val\n";
    }
  }
  eval { $spot = $levels[$level]->[$y]->[$x]; };
#  print "inserting \@ ($x, $y, $level)\n";
  if (defined $spot)
  { if ($objtype =~ /^(monster|player)/i) # monster or player, one per tile
    { if (defined $spot->[1])
      { $at = $spot->[1]->{'type'};
        if ($at =~ /^(monster|player)/i)
        { return 0; }
      }
      # always in slot one
      splice @$spot, 1,0, ($object);
    }
    else # not monster or player
    { push @$spot, $object;
    }   
    $object->{'level'} = $level;
    $object->{'x'} = $x;
    $object->{'y'} = $y;
    return 1;
  }
  0;
}

sub remove_object
{ my ($object, $level, $x, $y) = @_;
  my ($spot, $ll);
  eval { $spot = $levels[$level]->[$y]->[$x]; };
  if (defined $spot)
  { $ll = @$spot;
    for ($x = 1; $x <$ll; $x++)
    { if ($object == $spot->[$x])
      { splice @$spot, $x, 1; }
    }
  }
}

# move object, adjust lists
sub moveto
{ my ($object, $x, $y, $map) = @_;
  my ($oldx, $oldy, $oldlevel) = 
	($object->{'x'}, $object->{'y'}, $object->{'level'});
  if ( $object->{'startx'} <= 0 )
  { $object->{'startx'} = $object->{'x'};
    $object->{'starty'} = $object->{'y'};
    $object->{'startmap'} = $object->{'level'};
  }
  #print "moving to ($x, $y, $level)\n";
  if (&insert_object ( $object, $map, $x, $y))
  { &remove_object ( $object, $oldlevel, $oldx, $oldy);
    return (1);
  }
  0;
}

# abstract to notify player/object of stuff
sub notify
{ my ($object, $message, $sender) = @_;
  #return if ($object->{'type'} !~ /^(monster|player)/) ||
  return if (($object->{'type'} eq 'monster') && ($sender->{'type'} eq 'monster'));
  my($name);
  #eval { $name = $sender->{'name'} || $sender->{'species'}; };
  $name = $sender->{'name'} || $sender->{'species'};
  #$object->{'msg'} = [] if (! defined $object->{'msg'}); # safety trap
  $message = $name.$message if $message =~ /\w/; # attatch name only if a real message
#  while ($m = substr($message, 0, 40))
  { if ($message)
    { push @{$object->{'msg'}}, $message ;
      push @{$object->{'msgsend'}}, $sender;
#if (($object->{'type'} eq 'monster') && ($message eq ' '))
{ #print "monster note '$message'\n" ;
  #&callstack();
}
    }
#    $message = substr($message, 40);
  }
  $object->{'redraw'} = 1;
  while (@{$object->{'msg'}} > $maxmsg)
  { shift @{$object->{'msg'}};
    shift @{$object->{'msgsend'}};
  }
  #$sender->{'showmove'} = 1;
  &rampage ($object) if $object->{'type'} eq 'monster';
  $object->{'newmessage'} = 1;
}

# notify all matching patern in range
sub notifyrange
{ my ($sender, $message, $typespec) = @_;
  my ($x, $y, $level, $nx, $ny, $l, $ly);
  $level = $sender->{'level'};
  $l = $levels[$level];
  for ($y = 6; $y >= -6; $y--)
  { $ny = $y  + $sender->{'y'};
    next if ($ny < 0);
    $ly = $l->[$ny];
    next if (! ref $ly);
    for ($x = -6; $x <= 6; $x++)
    { $nx = $sender->{'x'}+$x;
      next if ($nx < 0);
      undef $spot;
      #$tile = '?';
      if (($x != 0) || ($y != 0))
      { #eval { $spot = $ly->[$nx]->[1];} ;
        $spot = $ly->[$nx]->[1];
#print "noise1..." if $message eq ' ';
	#if ((ref $spot) && ($spot->{'type'} =~ /^(monster|player)/i) )
	if ((ref $spot) && (($spot->{'type'} eq 'monster') || ($spot->{'type'} eq 'player')))
	{ &notify($spot, $message, $sender); }
      }
    } # for x
  } # for y
  $sender->{'showmove'} = 1;
} # sub notifyrange

sub notifyrangespeed
{ my ($sender, $message, $typespec) = @_;
  my ($x, $y, $level, $nx, $ny, $l, $ly);
  my (@gotten);
  $level = $sender->{'level'};
  $l = $levels[$level];
  for ($y = 6; $y >= -6; $y--)
  { $ny = $y  + $sender->{'y'};
    next if ($ny < 0);
    $ly = $l->[$ny];
    next if (! ref $ly);
    for ($x = -6; $x <= 6; $x++)
    { $nx = $sender->{'x'}+$x;
      next if ($nx < 0);
      undef $spot;
      #$tile = '?';
      if (($x != 0) || ($y != 0))
      { #eval { $spot = $ly->[$nx]->[1];} ;
        $spot = $ly->[$nx]->[1];
	if (ref $spot)
	{ push @gotten, $spot; }
#print '[';
      }
    } # for x
  } # for y
  foreach $spot (@gotten)
  { $spot->{'initiative'} += ((($spot->{'speed'})/($sender->{'speed'} || 1)) || 1);
#print "noise..." if $message eq ' ';
    while ($spot->{'initiative'} > 1)
    { &notify($spot, $message, $sender);
      $spot->{'initiative'} --;
#print "motion\n" if $message eq ' ';
    }
  }

  $sender->{'showmove'} = 1;
} # sub notifyrangespeed

sub gettile
{ my($at, $y, $level) = @_;
  my ($x, $spot, $move, $vis, $dig, $descrip);
  if (ref $at)
  { ($x, $y, $level) = ($at->{'x'}, $at->{'y'}, $at->{'level'}); }
  else
  { $x = $at; }
  return (' ', 0, 0, ' ') if (($x < 0) || ($y < 0));
  $spot = '';
  eval {$spot = $levels[$level]->[$y]->[$x]->[0]; };
  $move = $tiles{$spot}->{'move'};
  $vis = $tiles{$spot}->{'visable'};
  $dig = $tiles{$spot}->{'dig'};
  $descrip = $tiles{$spot}->{'descrip'};
  ($spot, $move, $vis, $dig, $descrip);
}

sub show_inventory
{ my ($player, $type, $exclude) = @_;
  my ($count, $shown, $obj, $name, %list, @list, %exclude);
  $type = $type || '.*';
  foreach $obj (@$exclude)  { $exclude{$obj} = 1; }
  foreach $obj (@{$player->{'inventory'}})
  { if (($obj->{'type'} =~ /$type/) && ($shown < 15) && (! $exclude{$obj}) )
    { $name = $obj->{'species'};
      if ($list{$name})
      { $list{$name}->{'count'} += ($obj->{'count'} || 1); }
      else
      { $list{$name} = {'count'=>($obj->{'count'} || 1), 'index'=>chr($count + ord('A'))};
        push @list, $name;
      }
      $shown ++;
#if ($type =~ /\|/)
#{ print "$type matching ",$obj->{'type'},"\n"; }
    }
    $count ++;
  }
  foreach $obj (@list)
  { 
    &notify ($player, $list{$obj}->{'index'}.">".$obj." - ".$list{$obj}->{'count'});
  }
  $shown;
}

# restore item to map in origional location after 'sensable' delay
sub respawn
{ my ($object) = @_;
  if ( $object->{'startx'} <= 0 )
  { $object->{'startx'} = $object->{'x'};
    $object->{'starty'} = $object->{'y'};
    $object->{'startmap'} = $object->{'level'};
  }
  if ($object->{'mode'})
  { return if (($object->{'rebirth'} > time()) || ( $object->{'doit'}));
    &insert_object($object,  $object->{'startmap'},
	$object->{'startx'}, $object->{'starty'});
    $object->{'mode'} = undef;
    $object {'respawned'} = 1;
  }
  else
  { $object->{'mode'} = \&respawn;
    $object->{'rebirth'} = time() + $object->{'respawn'};
    $object->{'count'} = $object->{'startcount'};
    $object {'respawned'} = 0;
    push @respawn, $object if ($object->{'type'} ne 'player');
  }
}

sub consume
{ my ($player, $item, $count) = @_;
  my ($indx, $q);
  $count = $count || 1;
  $item->{'count'} -= $count;
  if ($item->{'count'} <= 0)
  { &respawn ($item);
    for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
    { $q = $indx if ($player->{'inventory'}->[$indx] eq $item);
    }
    &dispose_object ($player, $q);
  }
  $item->{'count'};
}

# remove from inventory, unwear/wield as needed
sub dispose_object
{ my ($player, $index) = @_;
  my ($item) = splice (@{$player->{'inventory'}}, ($index), 1);

  foreach $stat (@wearplaces)
  { if ($item == $player->{'wear'}->{$stat})
    { $player->{'wear'}->{$stat} = undef;
      &stat_retotal($player);
    }
  }

  if ($item == $player->{'quiver'})
  { $player{'quiver'} = undef; }
}

# this never got finished. Was to provide extra protextion from bad magic
sub prep_magic
{ my ($item) = @_;
  if (! $item->{'magic'})
  { $item->{'magic'} = new Safe('mg'.$player->{'species'});
    $item->{'magic'}->permit_only (':base_core');
    $item->{'magic'}->share (&gettile, &moveto, &notify); # expand as may be needed
  }
  1;
}

# 'draw' projectile or thrown path
sub shoot_arrow
{ my ($dir, $dist, $sx, $sy, $stone) = @_;
  my ($x, $y, $dx, $dy, $at, $shoot);
  $stone = $stone || $arrows{$dir};
  $dx = $directions{$dir}->[0];   $dy = $directions{$dir}->[1];
  $x = $sx + 15;  $y = $sy + 12;
  while ($dist > 1)
  { $x += $dx;  $y -= $dy;
    $at = ansivt::home($x,$y);
    #push (@shoot, $at.$stone, $at.' ');
    $shoot .= $at.$stone;
    $dist --;
  }
  $shoot;
}

# scan inventory of object for item of type
sub object_has
{ my ($object, $type, $xtra) = @_;
  my ($i, $item);
  for ($i = 0; $i < @{$object->{'inventory'}} && !$item; $i ++)
  { $item = $object->{'inventory'}->[$i]
	if (($object->{'inventory'}->[$i]->{'species'} =~ /$type/) &&
	($object->{'inventory'}->[$i]->{'type'} =~ /$xtra/));
  }
  return ($item,$i) if $item;
  '';
}

# total stats dependent on stuff worn/wielded
@wearstats = ('AC', 'hit', 'lamp', 'vision', 'speed');
sub stat_retotal
{ my ($player) = @_;
  my ($place, $stat);
  foreach $stat (@wearstats)
  { $player->{$stat} = 0;
  }
  foreach $place (@wearplaces)
  { foreach $stat (@wearstats)
    { if ($stat eq 'AC')
      { $player->{$stat} += $player->{'wear'}->{$place}->{$stat}*(10-$player->{$stat})/10; }
      else { $player->{$stat} += $player->{'wear'}->{$place}->{$stat}; }
    }
  }
  foreach $stat (@wearstats)
  { $player->{$stat} += $player{"base-$stat"}; }
}

# some monsters carry treasure, droped when they're killed
sub spawn_treasure
{ my ($monster) = @_;
  my ($count, $treasure, $proto, $treasurecount, $key, @treasures, $item);
  $proto = $monsters{$monster->{'species'}};
  $treasurecount = rand (($proto->{'*treasurecount'}));
  @treasures = split ';', $proto->{'*treasures'};
  #print 'treasure',$monster->{'species'},';',$proto->{'*treasurecount'},"($treasurecount)",
#	$proto->{'*treasures'},"\n";
  if ($treasurecount > 0)
  { for ($count = 0; $count < $treasurecount; $count ++)
    { ($item, $ic) = split ',', $treasures[rand(@treasures)];
      #print "$ic $item\n";
      $item = $monsters{$item};
      next if (! ref $item);
      foreach $key (keys (%{$item}) )
      { $treasure->{$key} = $item->{$key} if ($key !~ /^[_|\*]/); }
      $treasure->{'count'} = $ic || 1;
      &insert_object($treasure, $monster->{'level'}, $monster->{'x'},$monster->{'y'});
      $treasure->{'norespawn'} = 1; # this object once droped or destroyed is gone
      #print "created ", $monster->{'level'},' ', $monster->{'x'},' ',$monster->{'y'},"\n";
    }
    #print "treasure spawned\n";
  }
}

sub revnumerically { $b <=> $a; }
sub object_dist
{ my ($m1, $m2) = @_;
#  &notify ($m1, "distancing...");
#  &notify ($m2, "distancing...");  
  # max (abs(m1->x - m2->x), abs(m1->y - m2->y)
  my ($max);
  ($max) = sort revnumerically abs($m1->{'x'} - $m2->{'x'}), abs($m1->{'y'} - $m2->{'y'});
#  &notify ($m1, "distancing...$max");
#  &notify ($m2, "distancing...$max");  
  $max;
}

sub loc_obj_dist
{ my ($x, $y, $m2) = @_;
  my ($max);
  ($max) = sort revnumerically abs($x - $m2->{'x'}), abs($y - $m2->{'y'});
  $max
}

sub inlist
{ my ($spot, $gotten) = @_;
  my ($chk);
  foreach $chk (@$gotten)
  { return 1 if ($chk == $spot); }
  0;
}

# debugging usefulls
sub breakpoint # call this to get a breakpoint inside a monster
{ print "monster break\n";
}
sub dumpobj
{ my ($obj) = @_;
  my ($key, $val);
  foreach $key (sort keys %{$obj})
  { $val = $obj->{$key};
    print "$key = $val\n";
  }
}
sub callstack
{ my ($key, $package, $filename, $line);
  for ($key = 5; $key >= 1; $key--)
  { ($package, $filename, $line) = caller ($key);
    print "$filename: $line\n";
  }
}
# ================
# process login
# ================
sub login
{ my ($player, $cmd) = @_;
  my ($line, $playstat, $name, $passwd, $key, $val);
select $player->{'socket'};
$| = 1; # this shouldn't be needed.
#print "reading login...";

  if ($line = &getstring($player, $cmd)) # =~ /(\w+)\s+(\w+)\n/)
  { $line =~ s/\s/ /g;
    $line =~ s/[^\w ]//g;
    ($name, $passwd) = split /\s/, $line;
    # print "checking $name\r\n";
    # check if login valid for player
    $playstat = $players{$name};
    # if valid, note as logged in, get stats from %players,
    if ($name && (ref $playstat) && defined($playstat->{'passwd'}) &&
	($playstat->{'passwd'} eq $passwd) && ($playstat->{'li'} ne 'in') )
    { $player->{'mode'} = undef;  # next loop processes as normal
      $playstat->{'li'} = 'li';
      print ansivt::cls(), "logged in, $name\n";
    # notify nearby players
      &notifyrange($player, $name.' connects');
      select STDOUT;
      print "login: $name\r\n";
      while ($cmd = getc $player->{'socket'}){} # clear stream buffer
      tnet::charmode ($player->{'socket'}, $player);
      #&gettermtype($player->{'socket'}, $player);
      # transfer stats from player master record to connection record
      foreach $key (keys %{$playstat})
      { $player->{$key} = $playstat->{$key}; }
    }
    # else, reset login, send new prompt.
    else
    { #print "$name : '", $playstat->{'passwd'}, "' : '", $playstat->{'name'},"'\n";
      print $loginprompt;
      select STDOUT;
    }
    $player->{'instring'} = "";
  }
}

# ==========================================
# if monster type has an AI, run it
# otherwise just stand there.
# ==========================================
sub rampage
{ my($monster) = @_;
  my ($code);
  my ($d, $direction, $close, $nc, $nx, $ny, $aq1, $aq2);
  my ($act, $actor, $acted);

  if ($monster->{'doit'})
  { eval { &{$monster->{'doit'}}($monster); }; }
  if ($monster->{'mode'})
  { eval { &{$monster->{'mode'}}($monster); }; }
  elsif (defined $monster->{'AI'})
  { #if ($monster->{'species'} eq 'merchant')
    #{ print "AI break\n"; }
    eval { &{$monster->{'AI'}}($monster); };
    print ($monster->{'species'},' : ',$@) if $@;
    &notifyrange($monster, ''); # signal for updates  
  }
  
  # generic hostile
  # rather stupid monsters
  elsif ($monster->{'type'} =~ /^(monster|player)/) 
  { my ($act, $actor, $direction, $acted, $dx, $dy);
    $aq1 = $monster->{'msg'};
    $aq2 = $monster->{'msgsend'};
    while ((@$aq1 + 0) > 0)
    { $act = pop @$aq1;
      $actor = pop @$aq2;
      next if ($act ne ' ');
#print "rampaging ...\n";
      if ((ref $actor) && ($actor->{'type'} =~ /player/) )# move towards 
      { # find direction
        if (!($monster->{'dumb'})) # a smarter monster, not easily blocked by walls
	{ $close = &object_dist($monster, $actor);
	  $direction = 5;
	  for ($d = 1; $d <= 9; $d++)
	  { $nx = $monster->{'x'}+$directions{$d}->[0];
	    $ny = $monster->{'y'}+$directions{$d}->[1];
	    ($spot, $move) = &gettile($nx, $ny, $monster->{'level'});
	    if (($move > 0) && (($nc = &loc_obj_dist($nx, $ny, $actor)) <= $close))
	    { $direction = $d;   $close = $nc; }
	  }
	  &movei ($monster, $direction); # 
	  $acted++;
	  $monster->{'move'} = $direction;
	  $monster->{'time'} = (time()+5); # has monsters attention
	}
	else # standard dumb monster
	{ $dx = $actor->{'x'} <=> $monster->{'x'};
          $dy = $actor->{'y'} <=> $monster->{'y'};
	  $direction = $revdir{"$dx,$dy"};
          # move towards, attack
	  &movei ($monster, $direction); 
	  $acted++;
	  $monster->{'move'} = $direction;
          #&notifyrange($monster, ''); # signal for updates
	  $monster->{'time'} = (time()+5); # has monsters attention
	}
      }
    }
    undef $monster->{'msg'};	# clear, but shouldn't be needed
    undef $monster->{'msgsend'};    
    if (($acted <= 0) && ($monster->{'time'} < time()) )# pick a direction and wander
    { $monster->{'move'} += rand();
      $monster->{'move'} = 0 if ($monster->{'move'} > 9);
      &movei ($monster, $monster->{'move'}+1);
      #&notifyrange($monster, ''); # signal for updates
      $monster->{'time'} = time();
#print "random\n";
    }
  }
} # sub rampage

# ============================================================
# command functions
# every game command has a function here
# direct operation on game structures should be minimal here
# ============================================================

sub help
{ my ($player, $cmd) = @_;
  my ($line);
  foreach $line (split "\n", $helpmain)
  { &notify ($player, $line); }
}

# just look up character for now
sub looksym
{ my ($player, $cmd) = @_;
  &notify ($player, "enter symbol> ");
  $player->{'mode'} = \&looksym2;
  return;
}

sub looksym2
{ my ($player, $cmd) = @_;
  my ($d, $m);
  if ($d = $tiles{$cmd}->{'descrip'}){}
  elsif ($cmd eq '@') { $d = 'player';}
  else # look up monster
  { foreach $m (values %monsters)
    { if ($cmd eq $m->{'symbol'})
      { $d = $m->{'species'};#$m->{'type'}. ' '. $m->{'species'};
      }
    }
  }
  $d = $d || '???';
  &notify ($player, $cmd." = ".$d);
  $player->{'mode'} = 0;
}

sub helpcmd
{ my ($player, $cmd) = @_;
  my ($key, $line, @keys);
  if ($player->{'mode'} == undef)
  { &notify ($player, "enter command> ");
    $player->{'mode'} = \&helpcmd;
    return;
  }
  else
  { $player->{'mode'} = undef;
    @keys = keys %help;
    foreach $key (@keys)
    { if ($cmd =~ /$key/)
      { foreach $line (split (/\n/, $help{$key}) )
        { &notify($player, $line); }
      }
    }
  }
} # sub helpcmd

# move based on command
sub movei
{ my ($player, $direction) = @_;
  my (@monster, $name);
  $direction = substr($direction, 0,1);
  return if $direction !~ /^[12346789ykuhlbjn]/;  # just in case something calls with bad parameter
#if ($player->{'species'} eq 'Troll')
#{ &callstack(); }
#print "moving ", $player->{'species'}, "\n";
  if ((@monster = &findstuff($player, $direction, 1) ) &&
	( defined $monster[0]) &&
	($monster[0]->{'type'} =~ /^(monster|player)/i) )
  { &fight ($player, $direction, $monster[0]); }
  else
  { my($tile, $move) = &gettile($player->{'x'}+$directions{$direction}->[0], 
	$player->{'y'}+$directions{$direction}->[1],
	$player->{'level'});
#print "moving into $tile, terrain $move\n";
    if (( rand (1) < $move*($player->{'speed'} || 1) ) ||
	($player->{'speed'} == -1))  # && # some terrain is hard to traverse
	# and some creatures are slower than standard
    { if (&moveto ($player, $player->{'x'}+$directions{$direction}->[0],
  	$player->{'y'}+$directions{$direction}->[1],
	$player->{'level'}))
    # notify player of top item in current tiles stack
      { @monster = &findstuff($player);
        $name = $monster[1]->{'type'}.' '.$monster[1]->{'species'};
        &notify ($player, "found $name") if (($name =~ /\w/) && ($name !~ /^(player|monster)/));
	#&notifyrangespeed($player, ' ');
      }
    }
    # else { &notifyrange($player, ' '); }
  }
  $player->{'redraw'} = 1;
}

# move, grabing direction from next key
# largely redundent, may never impliment
sub go
{ my ($player, $direction) = @_;

}

# attack monster, scaning direction if monster not given
sub fight
{ my ($player, $direction, $monster, $hit) = @_;
  my (@monster, $name, $damage, $XP);
  
  if ($direction eq 'F')
  { $player->{'mode'} = \&fight;
    #$noact = 1;
    return;
  }
  if (! (ref $monster))
  { if ((@monster = &findstuff($player, $direction, 1) ) &&
	( defined $monster[0]) &&
	($monster[0]->{'type'} =~ /^(monster|player)/i) )
    { $monster = $monster->[0]; }
  }
  if (! (ref $monster) )
  { &notify ($player, "fight who?"); $player->{'mode'} = undef; return; }
  if (($player->{'type'} eq 'player') && ($monster->{'type'} eq 'player') &&
	!$PvP)
  { $player->{'mode'} = undef; return; }
  $name = $monster->{'name'} || $monster->{'species'};
  $dammage = $hit || $player->{'hit'} || 10;
  if ($dammage < ($monster->{'health'}-$monster->{'dammage'}))  { $XP = $dammage; }
  else { $XP = $monster->{'health'} - $monster->{'dammage'}; } 
  $player->{'experience'} += $XP; # if $monster->{'health'};
#print "experience = ".$player->{'experience'}. "\n";
  $dammage *= (10-$monster->{'AC'})/10;
  $dammage *= (log($player->{'experience'} * $exp_bonus + 2.71828));
  $dammage = 0 if !$monster->{'health'};
  if ($dammage) # some monsters are special
  { 
    if ($monster->{'type'} eq 'player')
    { &notify ($monster, " has attacked you", $player);
      # automatic counter attack, ballances P v. P
      &fight ($monster, 0, $player) if $direction && ($player->{'type'} eq 'player');
    }
    $monster->{'dammage'} += $dammage;
    if ($monster->{'dammage'} > $monster->{'health'})
    { # tempory
      $monster->{'dammage'} = 0;
      $monster->{'experience'} = 0;
      &remove_object ($monster, $monster->{'level'}, $monster->{'x'}, $monster->{'y'});
      &spawn_treasure($monster);
      &respawn ($monster);
      &notify ($player, "$name killed");
      &notify ($monster, " has killed you", $player);
    }
    else
    { &notify ($player, "fighting $name...");
    }
  }
  else
  { #&notify ($player, "pushing $name...");
    &notify ($monster, " pushes", $player);
  }
  $player->{'mode'} = undef if ($player->{'mode'} == \&fight);
  #&notifyrangespeed($player, ' ');
}

# healing rest
sub rest
{ my ($player) = @_;
  $player->{'dammage'} -- if ($player->{'dammage'}> 0);
  $player->{'dammage'} = 0 if ($player->{'dammage'} < 0);
}

sub quit
{ my ($player, $command) = @_;
  my ($next) = getc($player->{'socket'});
  if ( (defined $next) && ($next == "\xff") && ($command ne '!!!') )
  { tnet::telnet($player->{'socket'}, $player);
    return;
  }
  # M$ telnet sometimes sends commands without IAC prefix
  if ( (ord($lastcommand) > 128) && ($command ne '!!!') )
  { #print STDOUT ord($lastcommand),'<last next>',ord($next), "$next\n";
    return;
  }

  select $player->{'socket'};
  print ansivt::home(2,21), "quitting...\r\n";
  select STDOUT;
  my ($name) = $player->{'name'};
  print "player $name quitting\n";
  &notifyrange($player, "$name exiting");
  vec($sockets, fileno($player->{'socket'}), 1) = 0;
  close $player->{'socket'};
  $player->{'quitting'} = 1; # flag for removal
  $player->{'li'} = ""; # no longer logged in;
  $player->{'mode'} = undef;
  # copy player stats back
  
  # remove from map
  &remove_object($player, $player->{'level'}, $player->{'x'}, $player->{'y'});
}

sub drop
{ my ($player, $cmd) = @_;
  my ($item);
  if ($player->{'mode'} == undef)
  { &notify ($player, "what to drop?>");
    $player->{'mode'} = \&drop;
    # list inventory
    my @exclude = values %{$player->{'wear'}}; # rather not drop what you're wearing
    if (! &show_inventory($player, undef, \@exclude))
    { &notify ($player, "nothing to drop");
      $player->{'mode'} = 0;
    }
    return;
  }
  else
  { if ((! ref $stuff[1]->{'drop'}) ||
	(eval {&{$stuff[1]->{'drop'}}($stuff[1], $player, $player);}))
    { # transfer named item, if it exists, to tile player is on
      $cmd = lc $cmd;
      $player->{'mode'} = undef;
      $item = $player->{'inventory'}->[ord($cmd) - ord('a')];
      if ( (ref $item) && (&insert_object($item, $player->{'level'}, $player->{'x'}, $player->{'y'})) )
      { &dispose_object($player, ord($cmd) - ord('a'));
        # say what happened
        &notify($player, "dropped ".$item->{'species'});
        &notifyrange($player, " has dropped ".$item->{'species'});
      }
    }
    else
    { &notify($player, "drop what?"); }
  }
}

sub eat
{ my ($player, $cmd) = @_;
  if ($player->{'mode'} == undef)
  { &notify($player, "select food>");
    $player->{'mode'} = \&eat;
    # list inventory
    if (! &show_inventory($player, 'food'))
    { &notify ($player, "nothing to eat");
      $player->{'mode'} = 0;
    }
    return;
  }
  else
  { $player->{'mode'} = undef;
    $cmd = lc $cmd;
    if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
    { $item = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
    if ((ref $item) && ($item->{'type'} =~ /food/))
    { $player->{'hunger'} -= $item->{'health'};
      $player->{'hunger'} = 0 if $player->{'hunger'} < 0;
      $player->{'dammage'} -= $item->{'health'};
      $player->{'dammage'} = 0 if $player->{'dammage'} < 0;
      &notify ($player. $item->{'species'}. " eaten");
#      $item->{'count'} --;
#      if ($item->{'count'}<= 0)
#      { &respawn ($item);
#	&dispose_object ($player, ord($cmd) - ord('a'));
#      }
      &consume($player, $item);
    }
    else
    { &notify($player, "Uck! why eat that?");
    }
  }
}

# shoot projectile in direction
sub fire
{ my ($player, $cmd) = @_;
  # get a direction
  if (! $player->{'quiver'})
  { &notify($player, "select a dart/arrow");
    return;
  }
  &notify ($player, "Aim...");
  $player->{'mode'} = \&fire1;
}
sub fire1
{ my ($player, $cmd) = @_;
  my (@targets, $indx, $bow, $range, $hit, $q);

  $player->{'mode'} = undef;
  if ($player->{'quiver'}->{'type'} !~ /project/)
  { $player->{'throw'} = $player->{'quiver'};
    &throw2($player, $cmd);
    return;
  }
  if ($cmd !~ /[12346789ykuhlbjn]/)
  { &notify ($player, 'cancel');
    return;
  }
  for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
  { $bow = $player->{'inventory'}->[$indx]
	if ($player->{'inventory'}->[$indx]->{'species'} eq
		$player->{'quiver'}->{'bow'});
  }
  # range = bow.range && bow.exists || arrow.range
  if (ref $bow) { $range = $bow->{'range'}; }
  else { $range = $player->{'quiver'}->{'range'}; }
  
  # scan in direction for target
  @targets = &findstuff ($player, $cmd, $range, '^(monster|player)', 1);
  $shot = $targets[0];
  
  $player->{'effects'} .= &shoot_arrow($cmd, $dist, 0,0); #$player->{'x'}, $player->{'y'});

  if (ref $shot) #->{'type'} =~ /monster|player/i)
  { # hit = bow.hit && bow.exists || arrow.hit
    if (ref $bow) { $hit = $bow->{'hit'}; }
    else { $hit = $player->{'quiver'}->{'hit'}; }
    &fight ($player, 0, $targets[0], $hit);
    $shot->{'effects'} .= &shoot_arrow($cmd, $dist,
	$shot->{'x'} - $player->{'y'}, $shot->{'x'} - $player->{'y'})
	if ($shot->{'type'} eq 'player');
  }
  # if no target, arrow hits wall
  else
  { &notify( $player, "miss."); }
#  $player->{'quiver'}->{'count'} --;
#  if ($player->{'quiver'}->{'count'} <= 0)
#  { &respawn ($player->{'quiver'});
#    for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
#    { $q = $indx if ($player->{'inventory'}->[$indx] eq $player->{'quiver'});
#    }
#    &dispose_object ($player, $q);
#  }
  if (!&consume ($player, $player->{'quiver'}))
  { $player->{'quiver'} = 0; }
}

sub get
{ my ($player, $cmd) = @_;
  my (@stuff, $indx, $oldstuff, $invitem, $dupflag);
  # get list on current tile
  @stuff = &findstuff($player, 0, 0);
  # we're stuff[0], transfer stuff[1]
  $oldstuff = '';
  if (ref $stuff[1])
  { $stuff[1]->{'count'} = $stuff[1]->{'count'} || 1;
    if ((! ref $stuff[1]->{'get'}) ||
	(eval {&{$stuff[1]->{'get'}}($stuff[1], $player, $player);}))
    { if ($stuff[1]->{'type'} eq 'gold')
      { $player->{'gold'} += ($stuff[1]->{'gold'}||1)*($stuff[1]->{'count'}||1);
        &remove_object($stuff[1], $player->{'level'}, $player->{'x'}, $player->{'y'});
      }
      elsif ($stuff[1]->{'norespawn'}) # merge into matching already in inventory, but not worn item
      { #for ($indx = 0; ($indx < @{$player->{'inventory'}}) && (ref $stuff[1]); $indx ++)
        foreach $invitem (@{$player->{'inventory'}})
	{ #$invitem = $player->{'inventory'}->[$indx];
#print $invitem->{'species'}, "\n";
	  if (($invitem->{'species'} eq $stuff[1]->{'species'}) &&
		( $invitem->{'norespawn'} ) &&
		( ! ((grep ($invitem == $_, values %{$player->{'wear'}}))+0 ) ) )
	  { $invitem->{'count'} += ($stuff[1]->{'count'}||1);
	    $dupflag = 1;
	    last;
print "got dup ",$invitem->{'species'}, "\n";
	  }
#print "end search\n";
	}
	if (! $dupflag)
	{ push @{$player->{'inventory'}}, $stuff[1];
#print "got new\n";
	}
# print "undurable get\n";
      }
      else
      { push @{$player->{'inventory'}}, $stuff[1]; }
      
      &remove_object($stuff[1], $player->{'level'}, $player->{'x'}, $player->{'y'});
      &notify($player, "got ".$stuff[1]->{'species'} || $stuff[1]->{'type'});
      &notifyrange($player, " gets ".$stuff[1]->{'species'});
    }
  }
  else
  { &notify($player, "nothing here to get"); }  
}

sub inventory
{ &show_inventory($player) || &notify ($player, "nothing");
}

# drink a potion, apply effects on player
sub quaff
{ my ($player, $cmd) = @_;
  &notify ($player, "Select potion");
  if (! &show_inventory($player, 'potion'))
  { &notify ($player, "no potions");
    return;
  }  
  $player->{'mode'} = \&quaff1;
}
sub quaff1
{ my ($target, $cmd) = @_;
  my ($item, $q);
  $player->{'mode'} = undef;
  $cmd = lc $cmd;
  if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $item = $target->{'inventory'}->[ord($cmd) - ord('a')]; }
  if (($item->{'type'} eq 'potion')) # && (&prep_magic ($item)) )
  { &notify ($target, "quaffing ".$item->{'species'});
    # my ($target, $item) in scope for spell?
    #$item->{'magic'}->reval($item->{'spell'});
    eval {&{$item->{'spell'}}($item, $target, $target);};
# recycle potion
#    &respawn ($item);
#    for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
#    { $q = $indx if ($player->{'inventory'}->[$indx] eq $item);
#    }
#    &dispose_object ($player, $q);
    &consume ($player, $item);
  }
  else
  { &notify ($target, "quaff what?"); }
}

sub zap
{ my ($player, $cmd) = @_;
  &notify ($player, "Select wand");
  if (! &show_inventory($player, 'wand') )
  { &notify ($player, "no wands");
    return;
  }
  $player->{'mode'} = \&zap1;
}

sub zap1
{ my ($player, $cmd) = @_;
  my ($item, $q);

  $cmd = lc $cmd;
  if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $item = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
  if (($item->{'type'} eq 'wand'))
  { &notify ($target, "Aim...".$item->{'species'});
    $player->{'mode'} = \&zap2;
    $player->{'wand'} = $item;
  }
  else
  { &notify ($player, "zap what?");
    $player->{'mode'} = undef;
  }
}

sub zap2
{ my ($player, $cmd) = @_;
    my (@targets, $indx, $wand, $eff, $q);

  $player->{'mode'} = undef;

  if ($cmd !~ /[12346789ykuhlbjn]/)
  { &notify ($player, 'cancel');
    return;
  }
  $wand = $player->{'wand'};
  # scan in direction for target
  @targets = &findstuff ($player, $cmd, 10, '^(monster|player)', 1);
  $shot = $targets[0];
  
  $eff = &shoot_arrow($cmd, $dist, 0,0,$wand->{'effect'});
  $player->{'effects'} .= ansivt::color1($wand->{'color'}).$eff.ansivt::clearcolor();

  if (ref $shot) #->{'type'} =~ /monster|player/i)
  { eval {&{$wand->{'spell'}}($wand, $targets[0], $player);};
    $shot->{'effects'} .= &shoot_arrow($cmd, $dist,
	$shot->{'x'} - $player->{'y'}, $shot->{'x'} - $player->{'y'}, $wand->{'effect'})
	if ($shot->{'type'} eq 'player');
  }
  # if no target, arrow hits wall
  else
  { &notify( $player, "miss."); }
#  $player->{'wand'}->{'count'} --;
#  if ($player->{'wand'}->{'count'} <= 0)
#  { &respawn ($player->{'wand'});
#    for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
#    { $q = $indx if ($player->{'inventory'}->[$indx] eq $player->{'wand'});
#    }
#    &dispose_object ($player, $q);
  &consume($player, $item);
  $player->{'wand'} = 0;
}

# select projectile to fire
sub quiver
{ my ($player) = @_;
  &notify ($player, "Select projectile");
  if (! &show_inventory($player, 'projectile'))
  { &notify ($player, "no projectiles");
    return;
  }
  $player->{'mode'} = \&quiver1;
}

sub quiver1
{ my ($player, $cmd) = @_;
  $player->{'mode'} = undef;
  $cmd = lc $cmd;
  if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $item = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
  $player->{'quiver'} = $item;
}

# read nearby sign or 'book' in hand
sub read
{ my ($player) = @_;
  &notify ($player, "Select 'book' or direction");
  &show_inventory($player, 'book');
  $player->{'mode'} = \&read1;
}
sub read1
{ my ($player) = @_;
  my ($book, $line);

  $player->{'mode'} = undef;

  if ($cmd =~ /[12346789ykuhlbjn]/)
  { # scan in direction for target
    ($book) = &findstuff ($player, $cmd, 1, 'book', 1);
  }
  elsif ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $book = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
  if ($book->{'text'})
  { foreach $line(split "\n", $book->{'text'})
    { &notify ($player, $line); }
    if ($book->{'spell'})
    { eval {&{$book->{'spell'}}($book, $player, $player);}; }
  }
}

# record stats for later recall
# also log out
# game master save will need to deal with inventory
sub save
{ my ($player) = @_;
  my ($stat, $val, $master);
  $master = players{$player->{'name'}};
  while (($stat, $val) = each %$player)
  { if (! ref ($val))
    { $master->{$stat} = $val; }
  }
  &quit($player);
}

# toss object in direction
sub throw
{ my ($player, $cmd) = @_;
  &notify ($player, "Select item");
  if (! &show_inventory($player, '.*'))
  { &notify ($player, "nothing to throw");
    return;
  }
  $player->{'mode'} = \&throw1;
}
# get object
sub throw1
{ my ($player, $cmd) = @_;
  my ($item, $q);
  #$player->{'mode'} = undef;
  $cmd = lc $cmd;
  if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { if ((! ref $stuff[1]->{'get'}) ||
	(eval {&{$stuff[1]->{'get'}}($stuff[1], $player, $player);}))
    { $item = $player->{'inventory'}->[ord($cmd) - ord('a')];
      &notify ($target, "Aim...".$item->{'species'});
      $player->{'mode'} = \&throw2;
      $player->{'wand'} = $item;
    }
  }
  else
  { &notify ($player, "throw what?");
    $player->{'mode'} = undef;
  }
}
# get direction and throw
sub throw2
{ my ($player, $cmd) = @_;
    my (@targets, $indx, $item, $dropx, $dropy);

  $player->{'mode'} = undef;

  if ($cmd !~ /[12346789ykuhlbjn]/)
  { &notify ($player, 'cancel');
    return;
  }
  $item = $player->{'wand'};
  # scan in direction for target
  @targets = &findstuff ($player, $cmd, 10, '^(monster|player)', 1);
  $shot = $targets[0];
  $eff = &shoot_arrow($cmd, $dist, 0,0,$item->{'symbol'});
  $player->{'effects'} .= ansivt::color1($item->{'color'}).$eff.ansivt::clearcolor();

  if (ref $shot) #->{'type'} =~ /monster|player/i)
  { &fight ($player, 0, $shot, $item->{'hit'} || 10);
    $shot->{'effects'} .= &shoot_arrow($cmd, $dist,
	$shot->{'x'} - $player->{'y'}, $shot->{'x'} - $player->{'y'})
	if ($shot->{'type'} eq 'player');
  }
  # if no target, item hits wall
  # drop item at destination
  $dropx = $player->{'x'} + ($dist-1) * $directions{$cmd}->[0];
  $dropy = $player->{'y'} + ($dist-1) * $directions{$cmd}->[1];
  
  if ( (ref $item) && (&insert_object($item, $player->{'level'}, $dropx, $dropy)) )
  { for ($indx = 0; $indx < @{$player->{'inventory'}}; $indx++)
    { $q = $indx if ($player->{'inventory'}->[$indx] eq $player->{'wand'});
    }
    &dispose_object ($player, $q);
    $player->{'wand'} = 0;
  }
}

@wearplaces = ('cloak', 'armor', 'helmet', 'amulet', 'ring1', 'ring2', 'weapon', 'shield', 'shoe');
@wearstate = ('wear', 'wear', 'wear', 'wear', 'wear', 'wear', 'wield', 'wield', 'wear');

sub wear
{ my ($player, $cmd) = @_;
  my @exclude = values %{$player->{'wear'}};
#  if ($player->{'mode'} == undef)
    &notify($player, "select >");
    $player->{'mode'} = \&wear1;
    # list inventory
    if (! &show_inventory($player, '(cloak|armor|weapon|ring|amulet)', \@exclude ) )
    { &notify ($player, "nothing to wear/wield");
      $player->{'mode'} = 0;
    }
#    return;
}

sub wear1
{ my ($player, $cmd) = @_;
  my ($item, $indx, $place, $ip, $matched, $got, %exclude);
  $player->{'mode'} = undef;
  foreach $obj (values %{$player->{'wear'}})  { $exclude{$obj} = 1; }
  # set, or clear, weapon selection
  $cmd = lc $cmd;
  if ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $item = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
  if (ref $item)
  { if (((! ref $stuff[1]->{'wear'}) ||
	(eval {&{$stuff[1]->{'wear'}}($stuff[1], $player, $player);})) &&
	(! $exclude{$item}))
    { for ($indx = 0; $indx < @wearplaces; $indx ++)
      { $place = $wearplaces[$indx];
        $ip = $item->{'place'} || $item->{'type'};
        if ( ($place =~ /$ip/) && ($player->{'wear'}->{$place}->{'species'} eq '') )
        { if (!$got)
	  { &notify($player, $wearstate[$indx]."ing ".$item->{'species'});
	    &notifyrange($player, ' '.$wearstate[$indx].'s '.$item->{'species'});
	    $got ++;
	  }
	  $player->{'wear'}->{$place} = $item;
	  &stat_retotal($player);
	  $matched ++;
	  last if ($item->{'placemulti'} <= 0);
	}
      } # for each place
    } # if can wear
  }
  if (!$matched) { &notify($player, 'cancel'); }
}

sub unwear
{ my ($player) = @_;
  my ($place, $species, $indx, $indxc);
  &notify($player, "select to remove>");
  $player->{'mode'} = \&remove1;
  for ($indx = 0; $indx < @wearplaces; $indx ++)
  { $indxc = chr($indx + ord('a'));
    $place = $wearplaces[$indx];
    $species = $player->{'wear'}->{$place}->{'species'};
    &notify ($player, $indxc.')'.$place.' '.$species)
	if $species ne ''; # $player->{'wear'}->{$place}; #$species;
  }
}

sub remove1
{ my ($player, $cmd) = @_;
  my ($item, $indx, $place);
  $player->{'mode'} = undef;
  $indx = (ord(lc $cmd) - ord('a'));
  $place = $wearplaces[$indx];
  if (($item = $player->{'wear'}->{$place}) && ($item->{'species'} ne ''))
  { if ((! ref $stuff[1]->{'remove'}) ||
	(eval {&{$stuff[1]->{'remove'}}($stuff[1], $player, $player);}))
    { $player->{'wear'}->{$place} = 0;
      &notify($player, "removing ".$place.' '.$item->{'species'});
      &notifyrange($player, " removes ".$item->{'species'});
      &stat_retotal($player);
    }
  }
  else
  { &notify ($player, 'cancel'); }
}

# look at nearby object
sub look
{ my ($player) = @_;
  &notify ($player, "from inventory or direction");
  &show_inventory($player, '.*');
  $player->{'mode'} = \&look1;
}
sub look1
{ my ($player) = @_;
  my ($book, $line);

  $player->{'mode'} = undef;

  if ($cmd =~ /[12346789ykuhlbjn]/)
  { # scan in direction for target
    ($book) = &findstuff ($player, $cmd, 1, 'book', 1);
  }
  elsif ((ord($cmd) >= ord('a')) && (ord($cmd) <= ord('z')))
  { $book = $player->{'inventory'}->[ord($cmd) - ord('a')]; }
  &notify ($player, $book->{'type'}.';'.$book->{'species'}.';'.$book->{'name'});
  foreach $line(split "\n", $book->{'descrip'})
  { &notify ($player, $line); }
}  

# this is in stat line, may never impliment
sub gold
{}

sub chat
{ my ($player, $cmd) = @_;
  my ($line, $pline);
  select $player->{'socket'};
  tnet::linemode($player->{'socket'});
  print ansivt::home(2,21), ansivt::clreol(), "> ";
  $player->{'mode'} = \&chat1;
  $player->{'time'} = time();
}
sub chat1
{ my ($player, $cmd) = @_;
  my ($line, $pline);
  if ($cmd eq "\e")
  { $player->{'mode'} = undef; }
  elsif (($line = &getstring($player, $cmd)) ne '0')
  { if ($player->{'time'} != time())
    { tnet::charmode($player->{'socket'});
      while ($pline = substr ($line,0,40))
      { &notifyrange ($player, ': '.$pline);
        &notify ($player, ': '.$pline, $player);
        $line = substr($line,50);
      }
      $player->{'mode'} = undef;
      getc ($player->{'socket'}); # clean annoying partner of EOL
    }
  }
}

sub who
{ my ($player) = @_;
  my ($p,$c);
  &notify ($player, "currently on:");
  foreach $p (@players)
  { &notify ($player, $p->{'name'});
    $c ++;
  }
}

sub page
{ my ($player) = @_;
  my ($p,$c);
  &notify ($player, "page to:");
  $player->{'playerlist'} = undef;
  foreach $p (@players)
  { &notify ($player, chr(ord('a') + $c).') '.$p->{'name'});
    push @{$player->{'playerlist'}}, $p->{'name'};
    $c ++;
  }
  $player->{'mode'} = \&page1;
  $nodrop_players ++;
}

sub page1
{ my ($player, $cmd) = @_;
  my ($line, $pline, $select, $indx);
  $player->{'select'} = ord($cmd) - ord('a');
  if (($select < 0) || ($select > @{$player->{'playerlist'}}))
  { &notify ($player, 'cancel');
    $player->{'mode'} = undef;
    return;
  }
  foreach $indx (@players)
  { $player->{'select'} = $indx if ($indx->{'name'} eq $player->{'playerlist'}->[$select]); }
  select $player->{'socket'};
  tnet::linemode($player->{'socket'});
  print ansivt::home(2,21), ansivt::clreol(), "> ";
  $player->{'mode'} = \&page2;
  $player->{'time'} = time();
  
}

sub page2
{ my ($player, $cmd) = @_;
  my ($line, $pline);
  if ($cmd eq "\e")
  { $player->{'mode'} = undef; }
  elsif (($line = &getstring($player, $cmd)) ne '0')
  { if ($player->{'time'} != time())
    { tnet::charmode($player->{'socket'});
      while ($pline = substr ($line,0,40))
      { &notify ($player->{'select'}, ' pages: '.$pline, $player);
        &notify ($player, 'paging: '.$player->{'playerlist'}->[$player->{'select'}].$pline);
        $line = substr($line,50);
      }
      $player->{'mode'} = undef;
      getc ($player->{'socket'}); # clean annoying partner of EOL
    }
  }
}

sub telnet
{ my ($player) = @_;
  net::telnet($player->{'socket'}, $player);
}

#sub shutdown
#{ $shutdown++; }
# ===================================================
# main loop
# ===================================================
print "starting main loop\n";
while (!$shutdown)
{ $setime = 100;
  $nfound = select ($insocks=$sockets, undef, undef, $setime);
  if ($nfound == -1)
  { print STDERR "select error $!" if (!$shutdown);
    next;
  
    if ($! == &EINTR) { next; }
    else { die "select error: $!"; }
  }
  elsif ($nfound eq '')
  { if ($lasttime == time())
    { sleep (1); }
    $lasttime = time();
    print STDERR "premature timeout ($setime)\n$!\n";
  }
  else
  {
  if (vec ($insocks, fileno($incomingsock), 1))
  { # setup active socket for new connection
    my ($newsock);
    print "incoming connection ...\n";
    $addr = accept ($newsock, $incomingsock) ||
    	print STDERR "accept error $!";
    # send prompt and create player object marked as logging in
    select $newsock; $| = 1;
    #print tnet::linemode(); # why does this obstruct the loginprompt?
    print ansivt::cls(); #, $loginprompt;
    fcntl($newsock, F_SETFL(), O_NONBLOCK());
    select STDOUT;
    $newconnect = {'socket'=>$newsock, 'mode'=>\&login, 'addr'=>$addr,
	'time'=>time()};
    #tie ($newconnect->{'mode'}, 'watchscaler', 'player mode');
    &gatekeeper($newconnect);
    push @players, $newconnect;
    vec ($sockets, fileno($newsock), 1) = 1;
  }
  
  #foreach $player (@players)
  # we need a position to remove player from list on logout
  $playercount = @players;
  for ($pl = 0; $pl < $playercount; $pl ++)
  { $player = $players[$pl];
    if (($player->{'quitting'}) && (! $nodrop_players))
    { splice @players, $pl, 1;
      # undef %$player;
      # login data now falls out of scope
    }
    elsif (vec ($insocks, fileno($player->{'socket'}), 1))
    { # get command, interpret
      &move($player);
      foreach $monster(@{$player->{'inventory'}})
      { &rampage ($monster) ; }
    }
  }
  } # if incomming data
  
  # put this before players?
#  foreach $monster (@monsters)
  for ($indx = 0; $indx < @respawn; $indx ++)
  { # have monster do it's thing
    $monster = $respawn[$indx];
    &rampage ($monster);
    if ($monster->{'respawned'})
    { splice @respawn, $indx, 1; }
  }

  
  # send updates to affected players
  foreach $player (@players)
  { if (($player->{'redraw'}) && ($player->{'noredraw'}<= 0)) { &clip_redraw ($player); }
    if ($player->{'newmessage'}) { &redraw_message($player); }
    if ((time() - $player->{'time'}) > $timeout)
    { # player has timed out.  save them and terminate connection
      &save ($player);
    }
  }
  # if it's been awhile since last save
  if ((time() - $lastsave) > $savetime)
  # save maps and player file.  DON'T overwrite what we read to start
  {
  
  }
}


# =========================================

__END__

data defination, internal

$savetime	How often to save
$starttime	when game was started
$lastsave	time of last save
$sockets	vector of open sockets
$incommingsock	server incomming socket
%commands	table of command sub pairs.
%wizard		commands for player with wizard stat

@players	list of logged in players
  $players	ref to hash of player stats
    'instring'	accumulator for string input
    'mode'	ref to sub for current mode, if applicable
    'socket'	connected socket
    'inventory'	hash of inventory
    etc.
     
%players	table of all valid players, indexed by character name
@monsters	list of all monsters.  same format as @players
%monsters	stats of defined monster, treasure types, portals.
		same format as %players
@levels		Map of the world
  @mapx		X axis
    @mapy	Y axis
      @contents	
        tile	wall, passage, etc
	objects	hashrefs: what players, monsters, 'treasure' are there.

data defination, external

master file
{
monsters=<filename>
players=<filename>
(map=<filename>)+
tiles=<filename>
}

map
{
size=<x>,<y>
[MAP]
<y lines of x characters>
<blank line>
monster <type> <x>,<y> <respawn delay> <extra stats>
}

players
{
# comment
[OBJECT]
name=<name>	# name must be first. order of others unimportant
type=player
passwd=<passwd>
x=<x>
y=<y>
health=<health>
%inventory
  <object>=<count>
/
.etc.
}

monsters
{
# comment
[OBJECT]
type=<type>	# must be first, index field
mr=<mr>
}

tiles
{
<tile>\t<movement>,<visibility>,<dig>\tdescription
}
