#!/usr/bin/perl
# sky.pl [size=<size>]
# variation on dungeon.pl with clouds in a field

#Copyright 2002 Peter Hanely, hanelyp@calweb.com

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$pi = 3.1415926;

sub set_tile
{ my ($x, $y, $c) = @_;
  $x = int $x;  $y = int $y;
  $line = $base[$y];
#print "set ($x, $y), '$c'\n";
  if (chkbnd($x, $y) ) # && (substr($line, $x, 1) ne $room))
  { substr($base[$y], $x, 1) = $c;
  }
}

sub chkbnd
{ my ($x, $y) = @_;
  ($x >=0) && ($x < $size) && ($y >= 0) && ($y < $size );
}

sub generate_cloud
{ my ($psg) = shift @_; # where to build
#  print '|';
#  print "\nbuilding $psg\n";

  $clear  = $spawning = 0;

    $cx = int rand ($size);
    $cy = int rand ($size);

  $trycount = 0;
  $clear = 1;
  $rad = 1.5 + rand(5);
  # check if area is clear
  #print "($x1,$y1)($x2,$y2)\n";
  for ($dir = 0; $dir <= 2*$pi; $dir += $pi/$rad)
  { for ($r = 0; $r <= ($rad+2); $r += 0.5)
    { $spot = substr($base[int $cy + $r * sin $dir],
		int $cx + $r * cos $dir, 1);
      if (&chkbnd(int ($cy + $r * sin $dir),int ($cx + $r * cos $dir)) && ($spot eq $cloud))
      { $clear = 0; }
    }
  }

#  print '/';
#  print "room ($cx,$cy)*$rad\n";
  # if so, build cloud, 
  if ($clear != 0)
  { print "cloud ($cx,$cy)\n" if $watchit;
    for ($dir = 0; $dir <= 2*$pi; $dir += $pi/$rad/3)
    { for ($r = 0; $r <= $rad; $r += 0.5)
      { $dx = $cx + $r * cos $dir;
        $dy = $cy + $r * sin $dir;
        &set_tile($dx, $dy, $cloud);
      }
    }
  }
}

$cloud = '|';
$room = '.';
$passage = '#';
# a few defaults
$size = 50;

select STDERR;
$| = 1;

print "reading arguments...\n";
while ($arg = shift @ARGV)
{ ($name, $val) = split '=', $arg, 2;
  if ($name eq "size")
  { $size = $val || $size; }
    elsif ($name eq 'up')
  { push @up, $val; }
  elsif ($name eq 'down')
  { push @down, $val; }
  elsif ($name eq'name')
  { $levelname=$val; }
    else # a monster
  { $monsters{"\t$name"} = "\t$val";
  }
}

if (!%monsters)
{ %monsters = ("monster\tBat"=>$size*$size/200, "\tquaffachino"=>$size*$size/1000);
} # a few basic hostiles

print "establishing ground...\n";  
# init dungeon ground
for ($i = 0; $i < $size; $i ++)
{ $bline .= $room;
}
$bline .= "\n";
for ($i = 0; $i < $size; $i ++)
{ push @base, $bline;
}

print "excavating";
# generate rooms and passages
for ($cnt = 0; $cnt < ($size*$size / 10); $cnt++)
{ &generate_cloud(); }

print "\npopulating...\n";
# populate with critters
foreach $monster(keys %monsters)
{ print "$monster\n";
  $count = $monsters{$monster};
  for ($i = 0; $i < $count;)
  { $x = int rand $size;
    $y = int rand $size;
    if (substr($base[$y], $x, 1) eq $room)
    { substr($base[$y], $x, 1) = $passage; # mark as occupied
      push @entities, "$monster\t$x,$y\t60\t$x,$y\n";
      $i ++;
    }
  }
} # each monster type

#stairs, up and down
foreach $up (@up)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tupstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCH:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x) .','.($y1+$y);  # so incoming stair can find location
	$okstair ++;
	last SEARCH;
      }
    }
  }
  if (!$okstair) { print STDERR "BAD STAIR!!!\n\n"; }
}
foreach $up (@down)
{ $x = $y = 0;
  while (substr($base[$y], $x, 1) ne $room)
  { $x = int rand $size;
    $y = int rand $size;
  }
  substr($base[$y], $x, 1) = $passage; # mark as occupied
  push @entities, "\tdownstair\t$x,$y\t0\t$x,$y\t$up,$levelname\n";
  $okstair=0;
SEARCHA:
  for ($x1 = -1; ($x1<=1); $x1++)
  { for ($y1 = -1; ($y1 <= 1); $y1++)
    { if (substr($base[$y1+$y], $x1+$x, 1) eq $room)
      { push @head, "$up=".($x1+$x).','.($y1+$y);  # so incoming stair can find location
        $okstair ++;
	last SEARCHA;
      }
    }
  }
}

select STDOUT;
$| = 1;
# output.  pipe into file
print 	"level=$levelname\nsize=$size,$size\n",
	"sky.pl, Peter Hanely\n",
	join ("\n", @head),
	"\n[MAP]\n",
	@base, "\n";
print "\n\n", @entities;
